import random
import re
import socket
from contextlib import suppress
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Set,
    Tuple,
    cast,
    TYPE_CHECKING,
)

from django.conf import settings
from django.core.cache.backends.base import DEFAULT_TIMEOUT, get_key_func
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from valkey import Valkey
from valkey.exceptions import ConnectionError, ResponseError, TimeoutError
from valkey.typing import AbsExpiryT, EncodableT, ExpiryT, KeyT, PatternT

from django_valkey import pool
from django_valkey.compressors.identity import IdentityCompressor
from django_valkey.exceptions import CompressorError, ConnectionInterrupted
from django_valkey.serializers.pickle import PickleSerializer
from django_valkey.util import CacheKey

if TYPE_CHECKING:
    from django_valkey.cache import ValkeyCache

_main_exceptions = (TimeoutError, ResponseError, ConnectionError, socket.timeout)

special_re = re.compile("([*?[])")


def glob_escape(s: str) -> str:
    return special_re.sub(r"[\1]", s)


class DefaultClient:
    def __init__(
        self,
        server: str | Iterable,
        params: Dict[str, Any],
        backend: "ValkeyCache",
    ) -> None:
        self._backend = backend
        self._server = server
        if not self._server:
            error_message = "Missing connections string"
            raise ImproperlyConfigured(error_message)
        if not isinstance(self._server, (list, tuple, set)):
            self._server = self._server.split(",")

        self._params = params

        self.reverse_key = get_key_func(
            params.get("REVERSE_KEY_FUNCTION")
            or "django_valkey.util.default_reverse_key"
        )

        self._clients: List[Valkey | Any | None] = [None] * len(self._server)
        self._options: dict = params.get("OPTIONS", {})
        self._replica_read_only = self._options.get("REPLICA_READ_ONLY", True)

        serializer_path = self._options.get(
            "SERIALIZER", "django_valkey.serializers.pickle.PickleSerializer"
        )
        serializer_cls = import_string(serializer_path)

        compressor_path = self._options.get(
            "COMPRESSOR", "django_valkey.compressors.identity.IdentityCompressor"
        )
        compressor_cls = import_string(compressor_path)

        self._serializer: PickleSerializer | Any = serializer_cls(options=self._options)
        self._compressor: IdentityCompressor | Any = compressor_cls(
            options=self._options
        )

        self.connection_factory = pool.get_connection_factory(options=self._options)

    def __contains__(self, key: KeyT) -> bool:
        return self.has_key(key)

    def _has_compression_enabled(self) -> bool:
        return (
            self._options.get(
                "COMPRESSOR", "django_valkey.compressors.identity.IdentityCompressor"
            )
            != "django_valkey.compressors.identity.IdentityCompressor"
        )

    def get_next_client_index(
        self, write: bool = True, tried: List[int] | None = None
    ) -> int:
        """
        Return a next index for read client. This function implements a default
        behavior for get a next read client for a replication setup.

        Overwrite this function if you want a specific
        behavior.
        """
        if write or len(self._server) == 1:
            return 0

        if tried is None:
            tried = []

        if tried and len(tried) < len(self._server):
            not_tried = [i for i in range(0, len(self._server)) if i not in tried]
            return random.choice(not_tried)

        return random.randint(1, len(self._server) - 1)

    def get_client(
        self,
        write: bool = True,
        tried: List[int] | None = None,
    ) -> Valkey | Any:
        """
        Method used for obtain a raw valkey client.

        This function is used by almost all cache backend
        operations for obtain a native valkey client/connection
        instance.
        """
        index = self.get_next_client_index(write=write, tried=tried)

        if self._clients[index] is None:
            self._clients[index] = self.connect(index)

        return self._clients[index]  # type:ignore

    def get_client_with_index(
        self,
        write: bool = True,
        tried: List[int] | None = None,
    ) -> Tuple[Valkey | Any, int]:
        """
        Method used for obtain a raw valkey client.

        This function is used by almost all cache backend
        operations for obtain a native valkey client/connection
        instance.
        """
        index = self.get_next_client_index(write=write, tried=tried)

        if self._clients[index] is None:
            self._clients[index] = self.connect(index)

        return self._clients[index], index  # type:ignore

    def connect(self, index: int = 0) -> Valkey | Any:
        """
        Given a connection index, returns a new raw valkey client/connection
        instance. Index is used for replication setups and indicates that
        connection string should be used. In normal setups, index is 0.
        """
        return self.connection_factory.connect(self._server[index])

    def disconnect(self, index: int = 0, client: Valkey | Any | None = None) -> None:
        """
        delegates the connection factory to disconnect the client
        """
        if client is None:
            client = self._clients[index]

        if client is not None:
            self.connection_factory.disconnect(client)

    def set(
        self,
        key: KeyT,
        value: EncodableT,
        timeout: int | float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
        nx: bool = False,
        xx: bool = False,
    ) -> bool:
        """
        Persist a value to the cache, and set an optional expiration time.

        Also supports optional nx parameter. If set to True - will use valkey
        setnx instead of set.
        """
        nkey = self.make_key(key, version=version)
        nvalue = self.encode(value)

        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        original_client = client
        tried: List[int] = []
        while True:
            try:
                if client is None:
                    client, index = self.get_client_with_index(write=True, tried=tried)

                if timeout is not None:
                    # Convert to milliseconds
                    timeout = int(timeout * 1000)

                    if timeout <= 0:
                        if nx:
                            # Using negative timeouts when nx is True should
                            # not expire (in our case delete) the value if it exists.
                            # Obviously expire not existent value is noop.
                            return not self.has_key(key, version=version, client=client)

                        # TODO: check if this is still valid
                        # valkey doesn't support negative timeouts in ex flags
                        # so it seems that it's better to just delete the key
                        # than to set it and then expire in a pipeline
                        return bool(self.delete(key, client=client, version=version))

                return bool(client.set(nkey, nvalue, nx=nx, px=timeout, xx=xx))
            except _main_exceptions as e:
                if (
                    not original_client
                    and not self._replica_read_only
                    and len(tried) < len(self._server)
                ):
                    tried.append(index)
                    client = None
                    continue
                raise ConnectionInterrupted(connection=client) from e

    def incr_version(
        self,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        """
        Adds delta to the cache version for the supplied key. Returns the
        new version.
        """

        if client is None:
            client = self.get_client(write=True)

        new_key, old_key, value, ttl, version = self._incr_version(
            key, delta, version, client
        )
        self.set(new_key, value, timeout=ttl, client=client)
        self.delete(old_key, client=client)
        return version + delta

    def _incr_version(
        self,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Tuple:
        if version is None:
            version = self._backend.version

        old_key = self.make_key(key, version)
        value = self.get(old_key, version=version, client=client)

        try:
            ttl = self.ttl(old_key, version=version, client=client)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if value is None:
            error_message = f"Key '{key!r}' not found"
            raise ValueError(error_message)

        if isinstance(key, CacheKey):
            new_key = self.make_key(key.original_key(), version=version + delta)
        else:
            new_key = self.make_key(key, version=version + delta)

        return new_key, old_key, value, ttl, version

    def add(
        self,
        key: KeyT,
        value: EncodableT,
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        """
        Add a value to the cache, failing if the key already exists.

        Returns ``True`` if the object was added, ``False`` if not.
        """
        return self.set(key, value, timeout, version=version, client=client, nx=True)

    def get(
        self,
        key: KeyT,
        default: Any | None = None,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Any:
        """
        Retrieve a value from the cache.

        Returns decoded value if key is found, the default if not.
        """
        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)

        try:
            value = client.get(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if value is None:
            return default

        return self.decode(value)

    def persist(
        self, key: KeyT, version: int | None = None, client: Valkey | Any | None = None
    ) -> bool:
        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)

        return client.persist(key)

    def expire(
        self,
        key: KeyT,
        timeout: ExpiryT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout  # type: ignore

        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)

        # for some strange reason mypy complains,
        # saying that timeout type is float | timedelta
        return client.expire(key, timeout)  # type: ignore

    def expire_at(
        self,
        key: KeyT,
        when: AbsExpiryT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        """
        Set an expiry flag on a ``key`` to ``when``, which can be represented
        as an integer indicating unix time or a Python datetime object.
        """
        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)

        return client.expireat(key, when)

    def pexpire(
        self,
        key: KeyT,
        timeout: ExpiryT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout  # type: ignore

        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)

        # TODO: see if the casting is necessary
        # for some strange reason mypy complains,
        # saying that timeout type is float | timedelta
        return bool(client.pexpire(key, timeout))  # type: ignore

    def pexpire_at(
        self,
        key: KeyT,
        when: AbsExpiryT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        """
        Set an expiry flag on a ``key`` to ``when``, which can be represented
        as an integer indicating unix time or a Python datetime object.
        """
        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)

        return bool(client.pexpireat(key, when))

    def get_lock(
        self,
        key: KeyT,
        version: int | None = None,
        timeout: float | None = None,
        sleep: float = 0.1,
        blocking_timeout: float | None = None,
        client: Valkey | Any | None = None,
        thread_local: bool = True,
    ):
        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)
        return client.lock(
            key,
            timeout=timeout,
            sleep=sleep,
            blocking_timeout=blocking_timeout,
            thread_local=thread_local,
        )

    # TODO: delete this in future releases
    lock = get_lock

    def delete(
        self,
        key: KeyT,
        version: int | None = None,
        prefix: str | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        """
        Remove a key from the cache.
        """
        if client is None:
            client = self.get_client(write=True)

        try:
            return client.delete(self.make_key(key, version=version, prefix=prefix))
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def delete_pattern(
        self,
        pattern: str,
        version: int | None = None,
        prefix: str | None = None,
        client: Valkey | Any | None = None,
        itersize: int | None = None,
    ) -> int:
        """
        Remove all keys matching pattern.
        """

        if client is None:
            client = self.get_client(write=True)

        pattern = self.make_pattern(pattern, version=version, prefix=prefix)

        try:
            count = 0
            pipeline = client.pipeline()

            for key in client.scan_iter(match=pattern, count=itersize):
                pipeline.delete(key)
                count += 1
            pipeline.execute()

            return count
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def delete_many(
        self,
        keys: Iterable[KeyT],
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        """
        Remove multiple keys at once.
        """
        keys = [self.make_key(k, version=version) for k in keys]

        if not keys:
            return 0

        if client is None:
            client = self.get_client(write=True)

        try:
            return client.delete(*keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def clear(self, client: Valkey | Any | None = None) -> None:
        """
        Flush all cache keys.
        """

        if client is None:
            client = self.get_client(write=True)

        try:
            client.flushdb()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def decode(self, value: EncodableT) -> Any:
        """
        Decode the given value.
        """
        try:
            value = int(value)
        except (ValueError, TypeError):
            # Handle little values, chosen to be not compressed
            with suppress(CompressorError):
                value = self._compressor.decompress(value)
            value = self._serializer.loads(value)
        return value

    def encode(self, value: EncodableT) -> bytes | int:
        """
        Encode the given value.
        """

        if isinstance(value, bool) or not isinstance(value, int):
            value = self._serializer.dumps(value)
            return self._compressor.compress(value)

        return value

    def _decode_iterable_result(
        self, result: Any, convert_to_set: bool = True
    ) -> List[Any] | Any | None:
        if result is None:
            return None
        if isinstance(result, list):
            if convert_to_set:
                return {self.decode(value) for value in result}
            return [self.decode(value) for value in result]
        return self.decode(result)

    def get_many(
        self,
        keys: Iterable[KeyT],
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> dict:
        """
        Retrieve many keys.
        """

        if client is None:
            client = self.get_client(write=False)

        if not keys:
            return {}

        recovered_data = {}

        map_keys = {self.make_key(k, version=version): k for k in keys}

        try:
            results = client.mget(*map_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        for key, value in zip(map_keys, results):
            if value is None:
                continue
            recovered_data[map_keys[key]] = self.decode(value)
        return recovered_data

    def set_many(
        self,
        data: Dict[KeyT, EncodableT],
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> None:
        """
        Set a bunch of values in the cache at once from a dict of key/value
        pairs. This is much more efficient than calling set() multiple times.

        If timeout is given, that timeout will be used for the key; otherwise
        the default cache timeout will be used.
        """
        if client is None:
            client = self.get_client(write=True)

        try:
            pipeline = client.pipeline()
            for key, value in data.items():
                self.set(key, value, timeout, version=version, client=pipeline)
            pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def _incr(
        self,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Valkey | Any | None = None,
        ignore_key_check: bool = False,
    ) -> int:
        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)

        try:
            try:
                # if key expired after exists check, then we get
                # key with wrong value and ttl -1.
                # use lua script for atomicity
                if not ignore_key_check:
                    lua = """
                    local exists = server.call('EXISTS', KEYS[1])
                    if (exists == 1) then
                        return server.call('INCRBY', KEYS[1], ARGV[1])
                    else return false end
                    """
                else:
                    lua = """
                    return server.call('INCRBY', KEYS[1], ARGV[1])
                    """
                value = client.eval(lua, 1, key, delta)
                if value is None:
                    error_message = f"Key '{key!r}' not found"
                    raise ValueError(error_message)
            except ResponseError as e:
                # if cached value or total value is greater than 64-bit signed
                # integer.
                # elif int is encoded. so valkey sees the data as string.
                # In these situations valkey will throw ResponseError

                # try to keep TTL of key
                timeout = self.ttl(key, version=version, client=client)

                # returns -2 if the key does not exist
                # means, that key have expired
                if timeout == -2:
                    error_message = f"Key '{key!r}' not found"
                    raise ValueError(error_message) from e
                value = self.get(key, version=version, client=client) + delta
                self.set(key, value, version=version, timeout=timeout, client=client)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        return value

    def incr(
        self,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Valkey | Any | None = None,
        ignore_key_check: bool = False,
    ) -> int:
        """
        Add delta to value in the cache. If the key does not exist, raise a
        ValueError exception. if ignore_key_check=True then the key will be
        created and set to the delta value by default.
        """
        return self._incr(
            key=key,
            delta=delta,
            version=version,
            client=client,
            ignore_key_check=ignore_key_check,
        )

    def decr(
        self,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        """
        Decreace delta to value in the cache. If the key does not exist, raise a
        ValueError exception.
        """
        return self._incr(key=key, delta=-delta, version=version, client=client)

    def ttl(
        self, key: KeyT, version: int | None = None, client: Valkey | Any | None = None
    ) -> int | None:
        """
        Executes TTL valkey command and return the "time-to-live" of specified key.
        If key is a non-volatile key, it returns None.
        """
        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)
        if not client.exists(key):
            return 0

        t = client.ttl(key)

        if t >= 0:
            return t
        if t == -2:
            return 0

        # Should never reach here
        return None

    def pttl(
        self, key: KeyT, version: int | None = None, client: Valkey | Any | None = None
    ) -> int | None:
        """
        Executes PTTL valkey command and return the "time-to-live" of specified key.
        If key is a non-volatile key, it returns None.
        """
        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)
        if not client.exists(key):
            return 0

        t = client.pttl(key)

        if t >= 0:
            return t
        if t == -2:
            return 0

        # Should never reach here
        return None

    def has_key(
        self, key: KeyT, version: int | None = None, client: Valkey | Any | None = None
    ) -> bool:
        """
        Test if key exists.
        """

        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)
        try:
            return client.exists(key) == 1
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def iter_keys(
        self,
        search: str,
        itersize: int | None = None,
        client: Valkey | Any | None = None,
        version: int | None = None,
    ) -> Iterator[str]:
        """
        Same as keys, but uses cursors
        for make memory efficient keys iteration.
        """

        if client is None:
            client = self.get_client(write=False)

        pattern = self.make_pattern(search, version=version)
        for item in client.scan_iter(match=pattern, count=itersize):
            yield self.reverse_key(item.decode())

    def keys(
        self,
        search: str,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> List[Any]:
        """
        Execute KEYS command and return matched results.
        Warning: this can return huge number of results, in
        this case, it strongly recommended use iter_keys
        for it.
        """

        if client is None:
            client = self.get_client(write=False)

        pattern = self.make_pattern(search, version=version)
        try:
            return [self.reverse_key(k.decode()) for k in client.keys(pattern)]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def make_key(
        self, key: KeyT, version: int | None = None, prefix: str | None = None
    ) -> KeyT:
        """Return key as a CacheKey instance so it has additional methods"""
        if isinstance(key, CacheKey):
            return key

        if prefix is None:
            prefix = self._backend.key_prefix

        if version is None:
            version = self._backend.version

        return CacheKey(self._backend.key_func(key, prefix, version))

    def make_pattern(
        self, pattern: str, version: int | None = None, prefix: str | None = None
    ) -> KeyT:
        if isinstance(pattern, CacheKey):
            return pattern

        if prefix is None:
            prefix = self._backend.key_prefix
        prefix = glob_escape(prefix)

        if version is None:
            version = self._backend.version
        version_str = glob_escape(str(version))

        return CacheKey(self._backend.key_func(pattern, prefix, version_str))

    def sadd(
        self,
        key: KeyT,
        *values: Any,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)
        encoded_values = [self.encode(value) for value in values]
        return client.sadd(key, *encoded_values)

    def scard(
        self,
        key: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)
        return client.scard(key)

    def sdiff(
        self,
        *keys: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set[Any]:
        if client is None:
            client = self.get_client(write=False)

        nkeys = [self.make_key(key, version=version) for key in keys]
        return {self.decode(value) for value in client.sdiff(*nkeys)}

    def sdiffstore(
        self,
        dest: KeyT,
        *keys: KeyT,
        version_dest: int | None = None,
        version_keys: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            client = self.get_client(write=True)

        dest = self.make_key(dest, version=version_dest)
        nkeys = [self.make_key(key, version=version_keys) for key in keys]
        return client.sdiffstore(dest, *nkeys)

    def sinter(
        self,
        *keys: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set[Any]:
        if client is None:
            client = self.get_client(write=False)

        nkeys = [self.make_key(key, version=version) for key in keys]
        return {self.decode(value) for value in client.sinter(*nkeys)}

    def sinterstore(
        self,
        dest: KeyT,
        *keys: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            client = self.get_client(write=True)

        dest = self.make_key(dest, version=version)
        nkeys = [self.make_key(key, version=version) for key in keys]
        return client.sinterstore(dest, *nkeys)

    def smismember(
        self,
        key: KeyT,
        *members: Any,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> List[bool]:
        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)
        encoded_members = [self.encode(member) for member in members]

        return [bool(value) for value in client.smismember(key, *encoded_members)]

    def sismember(
        self,
        key: KeyT,
        member: Any,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)
        member = self.encode(member)
        return bool(client.sismember(key, member))

    def smembers(
        self,
        key: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set[Any]:
        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)
        return {self.decode(value) for value in client.smembers(key)}

    def smove(
        self,
        source: KeyT,
        destination: KeyT,
        member: Any,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if client is None:
            client = self.get_client(write=True)

        source = self.make_key(source, version=version)
        destination = self.make_key(destination)
        member = self.encode(member)
        return client.smove(source, destination, member)

    def spop(
        self,
        key: KeyT,
        count: int | None = None,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set | Any:
        if client is None:
            client = self.get_client(write=True)

        nkey = self.make_key(key, version=version)
        result = client.spop(nkey, count)
        return self._decode_iterable_result(result)

    def srandmember(
        self,
        key: KeyT,
        count: int | None = None,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> List | Any:
        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)
        result = client.srandmember(key, count)
        return self._decode_iterable_result(result, convert_to_set=False)

    def srem(
        self,
        key: KeyT,
        *members: EncodableT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)
        nmembers = [self.encode(member) for member in members]
        return client.srem(key, *nmembers)

    def sscan(
        self,
        key: KeyT,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set[Any]:
        if self._has_compression_enabled() and match:
            err_msg = "Using match with compression is not supported."
            raise ValueError(err_msg)

        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)

        cursor, result = client.sscan(
            key,
            match=cast(PatternT, self.encode(match)) if match else None,
            count=count,
        )
        return {self.decode(value) for value in result}

    def sscan_iter(
        self,
        key: KeyT,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Iterator[Any]:
        if self._has_compression_enabled() and match:
            err_msg = "Using match with compression is not supported."
            raise ValueError(err_msg)

        if client is None:
            client = self.get_client(write=False)

        key = self.make_key(key, version=version)
        for value in client.sscan_iter(
            key,
            match=cast(PatternT, self.encode(match)) if match else None,
            count=count,
        ):
            yield self.decode(value)

    def sunion(
        self,
        *keys: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set[Any]:
        if client is None:
            client = self.get_client(write=False)

        nkeys = [self.make_key(key, version=version) for key in keys]
        return {self.decode(value) for value in client.sunion(*nkeys)}

    def sunionstore(
        self,
        destination: Any,
        *keys: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            client = self.get_client(write=True)

        destination = self.make_key(destination, version=version)
        encoded_keys = [self.make_key(key, version=version) for key in keys]
        return client.sunionstore(destination, *encoded_keys)

    def close(self) -> None:
        close_flag = self._options.get(
            "CLOSE_CONNECTION",
            getattr(settings, "DJANGO_VALKEY_CLOSE_CONNECTION", False),
        )
        if close_flag:
            self._close()

    def _close(self) -> None:
        """
        default implementation: Override in custom client
        """
        num_clients = len(self._clients)
        for idx in range(num_clients):
            self.disconnect(index=idx)
        self._clients = [None] * num_clients

    def touch(
        self,
        key: KeyT,
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        """
        Sets a new expiration for a key.
        """

        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        if client is None:
            client = self.get_client(write=True)

        key = self.make_key(key, version=version)
        if timeout is None:
            return bool(client.persist(key))

        # Convert to milliseconds
        timeout = int(timeout * 1000)
        return bool(client.pexpire(key, timeout))

    def hset(
        self,
        name: str,
        key: KeyT,
        value: EncodableT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        """
        Set the value of hash name at key to value.
        Returns the number of fields added to the hash.
        """
        if client is None:
            client = self.get_client(write=True)
        nkey = self.make_key(key, version=version)
        nvalue = self.encode(value)
        return client.hset(name, nkey, nvalue)

    def hdel(
        self,
        name: str,
        key: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        """
        Remove keys from hash name.
        Returns the number of fields deleted from the hash.
        """
        if client is None:
            client = self.get_client(write=True)
        nkey = self.make_key(key, version=version)
        return client.hdel(name, nkey)

    def hlen(
        self,
        name: str,
        client: Valkey | Any | None = None,
    ) -> int:
        """
        Return the number of items in hash name.
        """
        if client is None:
            client = self.get_client(write=False)
        return client.hlen(name)

    def hkeys(
        self,
        name: str,
        client: Valkey | Any | None = None,
    ) -> List[Any]:
        """
        Return a list of keys in hash name.
        """
        if client is None:
            client = self.get_client(write=False)
        try:
            return [self.reverse_key(k.decode()) for k in client.hkeys(name)]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def hexists(
        self,
        name: str,
        key: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        """
        Return True if key exists in hash name, else False.
        """
        if client is None:
            client = self.get_client(write=False)
        nkey = self.make_key(key, version=version)
        return client.hexists(name, nkey)



import bisect
import hashlib
from typing import Dict, Iterable, Iterator, List, Tuple


class HashRing:
    nodes: List[str] = []

    def __init__(self, nodes: Iterable[str] = (), replicas: int = 128) -> None:
        self.replicas: int = replicas
        self.ring: Dict[str, str] = {}
        self.sorted_keys: List[str] = []

        for node in nodes:
            self.add_node(node)

    def add_node(self, node: str) -> None:
        self.nodes.append(node)

        for x in range(self.replicas):
            _key = f"{node}:{x}"
            _hash = hashlib.sha256(_key.encode()).hexdigest()

            self.ring[_hash] = node
            self.sorted_keys.append(_hash)

        self.sorted_keys.sort()

    def remove_node(self, node: str) -> None:
        self.nodes.remove(node)
        for x in range(self.replicas):
            _hash = hashlib.sha256(f"{node}:{x}".encode()).hexdigest()
            del self.ring[_hash]
            self.sorted_keys.remove(_hash)

    def get_node(self, key: str) -> str | None:
        n, i = self.get_node_pos(key)
        return n

    def get_node_pos(self, key: str) -> Tuple[str, int] | Tuple[None, None]:
        if len(self.ring) == 0:
            return None, None

        _hash = hashlib.sha256(key.encode()).hexdigest()
        idx = bisect.bisect(self.sorted_keys, _hash)
        idx = min(idx - 1, (self.replicas * len(self.nodes)) - 1)
        return self.ring[self.sorted_keys[idx]], idx

    def iter_nodes(self, key: str) -> Iterator[Tuple[str, str] | Tuple[None, None]]:
        if len(self.ring) == 0:
            yield None, None

        node, pos = self.get_node_pos(key)
        for k in self.sorted_keys[pos:]:
            yield k, self.ring[k]

    def __call__(self, key: str) -> str | None:
        return self.get_node(key)



import logging
from typing import Any, TypeVar, Generic, Iterator, AsyncGenerator, Set

from django.conf import settings
from django.core.cache.backends.base import BaseCache
from django.utils.module_loading import import_string

Client = TypeVar("Client")
Backend = TypeVar("Backend")


class BaseValkeyCache(BaseCache, Generic[Client, Backend]):
    def __init__(self, server: str, params: dict[str, Any]) -> None:
        super().__init__(params)
        self._server = server
        self._params = params
        self._default_scan_itersize = getattr(
            settings, "DJANGO_VALKEY_SCAN_ITERSIZE", 10
        )

        options: dict = params.get("OPTIONS", {})
        self._client_cls = self.get_client_class()
        self._client = None

        self._ignore_exceptions = options.get(
            "IGNORE_EXCEPTIONS",
            getattr(settings, "DJANGO_VALKEY_IGNORE_EXCEPTIONS", False),
        )
        self._log_ignored_exceptions = options.get(
            "LOG_IGNORE_EXCEPTIONS",
            getattr(settings, "DJANGO_VALKEY_LOG_IGNORED_EXCEPTIONS", False),
        )
        self.logger = (
            logging.getLogger(getattr(settings, "DJANGO_VALKEY_LOGGER", __name__))
            if self._log_ignored_exceptions
            else None
        )

    def get_client_class(self) -> type[Client] | type:
        options = self._params.get("OPTIONS", {})
        _client_cls = options.get("CLIENT_CLASS", self.DEFAULT_CLIENT_CLASS)
        return import_string(_client_cls)

    @property
    def client(self) -> Client:
        """
        Lazy client connection property.
        """
        if self._client is None:
            self._client = self._client_cls(self._server, self._params, self)
        return self._client

    def persist(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    async def apersist(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    def expire(
        self, key, timeout, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    async def aexpire(
        self, key, timeout, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    def expire_at(
        self, key, when, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    async def aexpire_at(
        self, key, when, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    def pexpire(
        self, key, timeout, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    async def apexpire(
        self, key, timeout, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    def pexpire_at(
        self, key, when, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    async def apexpire_at(
        self, key, when, version: int | None = None, client: Backend | None = None
    ) -> bool:
        pass

    def get_lock(
        self,
        key,
        version: int | None = None,
        timeout: float | int | None = None,
        sleep: float = 0.1,
        blocking_timeout: float | None = None,
        client: Backend | None = None,
        thread_local: bool = True,
    ):
        pass

    async def aget_lock(
        self,
        key,
        version: int | None = None,
        timeout: float | int | None = None,
        sleep: float = 0.1,
        blocking_timeout: float | None = None,
        client: Backend | None = None,
        thread_local: bool = True,
    ):
        """
        As of now, valkey's `lock` method is a sync method
        """
        return self.get_lock()

    def delete_pattern(
        self,
        pattern: str,
        version: int | None = None,
        prefix: str | None = None,
        client: Backend | None = None,
        itersize: int | None = None,
    ) -> int:
        pass

    async def adelete_pattern(
        self,
        pattern: str,
        version: int | None = None,
        prefix: str | None = None,
        client: Backend | None = None,
        itersize: int | None = None,
    ) -> int:
        pass

    def decode(self, value) -> Any:
        pass

    async def adecode(self, value) -> Any:
        """
        Serializers only have sync methods
        """
        return self.decode(value)

    def ttl(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> int | None:
        pass

    async def attl(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    def pttl(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    async def apttl(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    def iter_keys(
        self,
        search,
        itersize: int | None = None,
        client: Backend | None = None,
        version: int | None = None,
    ) -> Iterator:
        pass

    async def aiter_keys(
        self,
        search,
        itersize: int | None = None,
        client: Backend | None = None,
        version: int | None = None,
    ) -> AsyncGenerator:
        pass

    def keys(
        self, search: str, version: int | None = None, client: Backend | None = None
    ) -> list[Any]:
        pass

    async def akeys(
        self, search: str, version: int | None = None, client: Backend | None = None
    ) -> list[Any]:
        pass

    def make_key(self, key, version: int | None = None, prefix: str | None = None):
        pass

    async def amake_key(
        self, key, version: int | None = None, prefix: str | None = None
    ):
        return self.make_key(key, version, prefix)

    def sadd(
        self,
        key,
        *values: Any,
        version: int | None = None,
        client: Backend | None = None,
    ) -> int:
        pass

    async def asadd(
        self, *values: Any, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    def scard(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    async def ascard(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    def sdiff(
        self, *keys, version: int | None = None, client: Backend | None = None
    ) -> Set[Any]:
        pass

    async def asdiff(
        self, *keys, version: int | None = None, client: Backend | None = None
    ) -> Set[Any]:
        pass

    def sdiffstore(
        self,
        dest,
        *keys,
        version_dest: int | None = None,
        version_keys: int | None = None,
        client: Backend | None = None,
    ) -> int:
        pass

    async def asdiffstore(
        self,
        dest,
        *keys,
        version_dest: int | None = None,
        version_keys: int | None = None,
        client: Backend | None = None,
    ) -> int:
        pass

    def sinter(
        self, *keys, version: int | None = None, client: Backend | None = None
    ) -> Set[Any]:
        pass

    async def asinter(
        self, *keys, version: int | None = None, client: Backend | None = None
    ) -> Set[Any]:
        pass

    def sinterstore(
        self, dest, *keys, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    async def asintstore(
        self, dest, *keys, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    def smismember(
        self,
        key,
        *members: Any,
        version: int | None = None,
        client: Backend | None = None,
    ) -> list[bool]:
        pass

    async def asmismember(
        self,
        key,
        *members: Any,
        version: int | None = None,
        client: Backend | None = None,
    ) -> list[bool]:
        pass

    def sismember(
        self,
        key,
        member: Any,
        version: int | None = None,
        client: Backend | None = None,
    ) -> bool:
        pass

    async def asismember(
        self,
        key,
        member: Any,
        version: int | None = None,
        client: Backend | None = None,
    ) -> bool:
        pass

    def smembers(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> Set[Any]:
        pass

    async def asmembers(
        self, key, version: int | None = None, client: Backend | None = None
    ) -> Set[Any]:
        pass

    def smove(
        self,
        source,
        destination,
        member: Any,
        version: int | None = None,
        client: Backend | None = None,
    ) -> bool:
        pass

    async def asmove(
        self,
        source,
        destination,
        member: Any,
        version: int | None = None,
        client: Backend | None = None,
    ) -> bool:
        pass

    def spop(
        self,
        key,
        count: int | None = None,
        version: int | None = None,
        client: Backend | None = None,
    ) -> Set | Any:
        pass

    async def aspop(
        self,
        key,
        count: int | None = None,
        version: int | None = None,
        client: Backend | None = None,
    ) -> Set | Any:
        pass

    def srandmember(
        self,
        key,
        count: int | None = None,
        version: int | None = None,
        client: Backend | None = None,
    ) -> list | Any:
        pass

    async def asrandmember(
        self,
        key,
        count: int | None = None,
        version: int | None = None,
        client: Backend | None = None,
    ) -> list | Any:
        pass

    def srem(
        self, key, *members, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    async def asrem(
        self, key, *members, version: int | None = None, client: Backend | None = None
    ) -> int:
        pass

    def sscan(
        self,
        key,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Backend | None = None,
    ) -> Set[Any]:
        pass

    async def asscan(
        self,
        key,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Backend | None = None,
    ) -> Set[Any]:
        pass

    def sscan_iter(
        self,
        key,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Backend | None = None,
    ) -> Iterator:
        pass

    async def asscan_iter(
        self,
        key,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Backend | None = None,
    ) -> AsyncGenerator:
        pass

    def sunion(
        self, *keys, version: int | None = None, client: Backend | None = None
    ) -> Set[Any]:
        pass

    async def asunion(
        self, *keys, version: int | None = None, client: Backend | None = None
    ) -> Set[Any]:
        pass

    def sunionstore(
        self,
        destination: Any,
        *keys,
        version: int | None = None,
        client: Backend | None = None,
    ) -> int:
        pass

    async def asunionstore(
        self,
        destination: Any,
        *keys,
        version: int | None = None,
        client: Backend | None = None,
    ) -> int:
        pass

    def hset(
        self,
        name: str,
        key,
        value,
        version: int | None = None,
        client: Backend | None = None,
    ) -> int:
        pass

    async def ahset(
        self,
        name: str,
        key,
        value,
        version: int | None = None,
        client: Backend | None = None,
    ) -> int:
        pass

    def hdel(
        self,
        name: str,
        key,
        version: int | None = None,
        client: Backend | None = None,
    ) -> int:
        pass

    async def ahdel(
        self,
        name: str,
        key,
        version: int | None = None,
        client: Backend | None = None,
    ) -> int:
        pass

    def hlen(self, name: str, client: Backend | None = None) -> int:
        pass

    async def ahlen(self, name: str, client: Backend | None = None) -> int:
        pass

    def hkeys(self, name: str, client: Backend | None = None) -> list[Any]:
        pass

    async def ahkeys(self, name: str, client: Backend | None = None) -> list[Any]:
        pass

    def hexists(
        self,
        name: str,
        key,
        version: int | None = None,
        client: Backend | None = None,
    ) -> bool:
        pass

    async def ahexists(
        self,
        name: str,
        key,
        version: int | None = None,
        client: Backend | None = None,
    ) -> bool:
        pass



from typing import Any
from urllib.parse import parse_qs, urlparse

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from valkey import Valkey
from valkey.connection import ConnectionPool, DefaultParser
from valkey.sentinel import Sentinel
from valkey._parsers.url_parser import to_bool

from django_valkey.base_pool import BaseConnectionPool


class ConnectionFactory(BaseConnectionPool[Valkey, ConnectionPool]):
    path_pool_cls = "valkey.connection.ConnectionPool"
    path_base_cls = "valkey.client.Valkey"

    def disconnect(self, connection: type[Valkey] | type) -> None:
        """
        Given a not null client connection it disconnects from the Valkey server.

        The default implementation uses a pool to hold connections.
        """
        connection.connection_pool.disconnect()

    def get_parser_cls(self) -> type[DefaultParser] | type:
        cls = self.options.get("PARSER_CLASS", None)
        if cls is None:
            return DefaultParser
        return import_string(cls)


class SentinelConnectionFactory(ConnectionFactory):
    def __init__(self, options: dict):
        # allow overriding the default SentinelConnectionPool class
        options.setdefault(
            "CONNECTION_POOL_CLASS", "valkey.sentinel.SentinelConnectionPool"
        )
        super().__init__(options)

        sentinels = options.get("SENTINELS")
        if not sentinels:
            error_message = "SENTINELS must be provided as a list of (host, port)."
            raise ImproperlyConfigured(error_message)

        # provide the connection pool kwargs to the sentinel in case it
        # needs to use the socket options for the sentinels themselves
        connection_kwargs = self.make_connection_params(None)
        connection_kwargs.pop("url")
        connection_kwargs.update(self.pool_cls_kwargs)
        self._sentinel = Sentinel(
            sentinels,
            sentinel_kwargs=options.get("SENTINEL_KWARGS"),
            **connection_kwargs,
        )

    def get_connection_pool(self, params: dict) -> ConnectionPool | Any:
        """
        Given a connection parameters, return a new sentinel connection pool
        for them.
        """
        url = urlparse(params["url"])

        # explicitly set service_name and sentinel_manager for the
        # SentinelConnectionPool constructor since will be called by from_url
        cp_params = params
        cp_params.update(service_name=url.hostname, sentinel_manager=self._sentinel)
        pool = super().get_connection_pool(cp_params)

        # convert "is_master" to a boolean if set on the URL, otherwise if not
        # provided it defaults to True.
        is_master: list[str] = parse_qs(url.query).get("is_master")
        if is_master:
            pool.is_master = to_bool(is_master[0])

        return pool


def get_connection_factory(
    path: str | None = None, options: dict | None = None
) -> ConnectionFactory | SentinelConnectionFactory | Any:
    if path is None:
        path = getattr(
            settings,
            "DJANGO_VALKEY_CONNECTION_FACTORY",
            "django_valkey.pool.ConnectionFactory",
        )
    opt_conn_factory = options.get("CONNECTION_FACTORY")
    if opt_conn_factory:
        path = opt_conn_factory

    cls = import_string(path)
    return cls(options)



import random
import socket
import time
from collections import OrderedDict
from typing import Tuple, Any

from django.conf import settings
from valkey import Valkey
from valkey.exceptions import ConnectionError, ResponseError, TimeoutError
from valkey.typing import KeyT, EncodableT

from django_valkey.client.default import DEFAULT_TIMEOUT, DefaultClient
from django_valkey.exceptions import ConnectionInterrupted

_main_exceptions = (ConnectionError, ResponseError, TimeoutError, socket.timeout)


class Marker:
    """
    Dummy class for use as
    marker for herded keys.
    """

    pass


def _is_expired(x, herd_timeout: int) -> bool:
    if x >= herd_timeout:
        return True
    val = x + random.randint(1, herd_timeout)

    if val >= herd_timeout:
        return True
    return False


class HerdClient(DefaultClient):
    def __init__(self, *args, **kwargs):
        self._marker = Marker()
        self._herd_timeout: int = getattr(settings, "CACHE_HERD_TIMEOUT", 60)
        super().__init__(*args, **kwargs)

    def _pack(self, value: Any, timeout) -> Tuple[Marker, Any, int]:
        herd_timeout = (timeout or self._backend.default_timeout) + int(time.time())
        return self._marker, value, herd_timeout

    def _unpack(self, value: Tuple[Marker, Any, int]) -> Tuple[Any, bool]:
        try:
            marker, unpacked, herd_timeout = value
        except (ValueError, TypeError):
            return value, False

        if not isinstance(marker, Marker):
            return value, False

        now = time.time()
        if herd_timeout < now:
            x = now - herd_timeout
            return unpacked, _is_expired(x, self._herd_timeout)

        return unpacked, False

    def set(
        self,
        key: KeyT,
        value: EncodableT,
        timeout: int | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | None = None,
        nx: bool = False,
        xx: bool = False,
    ):
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        if timeout is None or timeout <= 0:
            return super().set(
                key,
                value,
                timeout=timeout,
                version=version,
                client=client,
                nx=nx,
                xx=xx,
            )

        packed = self._pack(value, timeout)
        real_timeout = timeout + self._herd_timeout

        return super().set(
            key,
            packed,
            timeout=real_timeout,
            version=version,
            client=client,
            nx=nx,
            xx=xx,
        )

    def get(self, key, default=None, version=None, client=None):
        packed = super().get(key, default=default, version=version, client=client)
        val, refresh = self._unpack(packed)

        if refresh:
            return default

        return val

    def get_many(self, keys, version=None, client=None):
        if client is None:
            client = self.get_client(write=False)

        if not keys:
            return {}

        recovered_data = OrderedDict()

        new_keys = [self.make_key(key, version=version) for key in keys]
        map_keys = dict(zip(new_keys, keys))

        try:
            results = client.mget(*new_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        for key, value in zip(new_keys, results):
            if value is None:
                continue

            val, refresh = self._unpack(self.decode(value))
            recovered_data[map_keys[key]] = None if refresh else val

        return recovered_data

    def set_many(
        self, data, timeout=DEFAULT_TIMEOUT, version=None, client=None, herd=True
    ):
        """
        Set a bunch of values in the cache at once from a dict of key/value
        pairs. This is much more efficient than calling set() multiple times.

        If timeout is given, that timeout will be used for the key; otherwise
        the default cache timeout will be used.
        """
        if client is None:
            client = self.get_client(write=True)

        set_function = self.set if herd else super().set

        try:
            pipeline = client.pipeline()
            for key, value in data.items():
                set_function(key, value, timeout, version=version, client=pipeline)
            pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def incr(self, *args, **kwargs):
        raise NotImplementedError

    def decr(self, *args, **kwargs):
        raise NotImplementedError

    def touch(self, key, timeout=DEFAULT_TIMEOUT, version=None, client=None):
        if client is None:
            client = self.get_client(write=True)

        value = self.get(key, version=version, client=client)
        if value is None:
            return False

        self.set(key, value, timeout=timeout, version=version, client=client)
        return True



from django_valkey.client.default import DefaultClient
from django_valkey.client.herd import HerdClient
from django_valkey.client.sentinel import SentinelClient
from django_valkey.client.sharded import ShardClient

__all__ = ["DefaultClient", "HerdClient", "SentinelClient", "ShardClient"]



import functools
from inspect import iscoroutinefunction
from typing import Any, Callable

from valkey import Valkey

from django_valkey.base import BaseValkeyCache
from django_valkey.client import DefaultClient
from django_valkey.exceptions import ConnectionInterrupted

CONNECTION_INTERRUPTED = object()


def omit_exception(method: Callable | None = None, return_value: Any | None = None):
    """
    Simple decorator that intercepts connection
    errors and ignores these if settings specify this.
    """

    if method is None:
        return functools.partial(omit_exception, return_value=return_value)

    @functools.wraps(method)
    def _decorator(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except ConnectionInterrupted as e:
            if self._ignore_exceptions:
                if self._log_ignored_exceptions:
                    self.logger.exception("Exception ignored")

                return return_value
            raise e.__cause__  # noqa: B904

    @functools.wraps(method)
    async def _async_decorator(self, *args, **kwargs):
        try:
            return await method(self, *args, **kwargs)
        except ConnectionInterrupted as e:
            if self._ignore_exceptions:
                if self._log_ignored_exceptions:
                    self.logger.exception("Exception ignored")
                return return_value
            raise e.__cause__

    return _async_decorator if iscoroutinefunction(method) else _decorator


class ValkeyCache(BaseValkeyCache[DefaultClient, Valkey]):
    DEFAULT_CLIENT_CLASS = "django_valkey.client.DefaultClient"

    @omit_exception
    def set(self, *args, **kwargs):
        return self.client.set(*args, **kwargs)

    @omit_exception
    def incr_version(self, *args, **kwargs):
        return self.client.incr_version(*args, **kwargs)

    @omit_exception
    def add(self, *args, **kwargs):
        return self.client.add(*args, **kwargs)

    def get(self, key, default=None, version=None, client=None):
        value = self._get(key, default, version, client)
        if value is CONNECTION_INTERRUPTED:
            value = default
        return value

    @omit_exception(return_value=CONNECTION_INTERRUPTED)
    def _get(self, key, default, version, client):
        return self.client.get(key, default=default, version=version, client=client)

    @omit_exception
    def delete(self, *args, **kwargs):
        result = self.client.delete(*args, **kwargs)
        return bool(result)

    @omit_exception
    def delete_pattern(self, *args, **kwargs):
        kwargs.setdefault("itersize", self._default_scan_itersize)
        return self.client.delete_pattern(*args, **kwargs)

    @omit_exception
    def delete_many(self, *args, **kwargs):
        return self.client.delete_many(*args, **kwargs)

    @omit_exception
    def clear(self):
        return self.client.clear()

    @omit_exception(return_value={})
    def get_many(self, *args, **kwargs):
        return self.client.get_many(*args, **kwargs)

    @omit_exception
    def set_many(self, *args, **kwargs):
        return self.client.set_many(*args, **kwargs)

    @omit_exception
    def incr(self, *args, **kwargs):
        return self.client.incr(*args, **kwargs)

    @omit_exception
    def decr(self, *args, **kwargs):
        return self.client.decr(*args, **kwargs)

    @omit_exception
    def has_key(self, *args, **kwargs):
        return self.client.has_key(*args, **kwargs)

    @omit_exception
    def keys(self, *args, **kwargs):
        return self.client.keys(*args, **kwargs)

    @omit_exception
    def iter_keys(self, *args, **kwargs):
        return self.client.iter_keys(*args, **kwargs)

    @omit_exception
    def ttl(self, *args, **kwargs):
        return self.client.ttl(*args, **kwargs)

    @omit_exception
    def pttl(self, *args, **kwargs):
        return self.client.pttl(*args, **kwargs)

    @omit_exception
    def persist(self, *args, **kwargs):
        return self.client.persist(*args, **kwargs)

    @omit_exception
    def expire(self, *args, **kwargs):
        return self.client.expire(*args, **kwargs)

    @omit_exception
    def expire_at(self, *args, **kwargs):
        return self.client.expire_at(*args, **kwargs)

    @omit_exception
    def pexpire(self, *args, **kwargs):
        return self.client.pexpire(*args, **kwargs)

    @omit_exception
    def pexpire_at(self, *args, **kwargs):
        return self.client.pexpire_at(*args, **kwargs)

    @omit_exception
    def lock(self, *args, **kwargs):
        return self.client.lock(*args, **kwargs)

    @omit_exception
    def close(self):
        self.client.close()

    @omit_exception
    def touch(self, *args, **kwargs):
        return self.client.touch(*args, **kwargs)

    @omit_exception
    def sadd(self, *args, **kwargs):
        return self.client.sadd(*args, **kwargs)

    @omit_exception
    def scard(self, *args, **kwargs):
        return self.client.scard(*args, **kwargs)

    @omit_exception
    def sdiff(self, *args, **kwargs):
        return self.client.sdiff(*args, **kwargs)

    @omit_exception
    def sdiffstore(self, *args, **kwargs):
        return self.client.sdiffstore(*args, **kwargs)

    @omit_exception
    def sinter(self, *args, **kwargs):
        return self.client.sinter(*args, **kwargs)

    @omit_exception
    def sinterstore(self, *args, **kwargs):
        return self.client.sinterstore(*args, **kwargs)

    @omit_exception
    def sismember(self, *args, **kwargs):
        return self.client.sismember(*args, **kwargs)

    @omit_exception
    def smembers(self, *args, **kwargs):
        return self.client.smembers(*args, **kwargs)

    @omit_exception
    def smove(self, *args, **kwargs):
        return self.client.smove(*args, **kwargs)

    @omit_exception
    def spop(self, *args, **kwargs):
        return self.client.spop(*args, **kwargs)

    @omit_exception
    def srandmember(self, *args, **kwargs):
        return self.client.srandmember(*args, **kwargs)

    @omit_exception
    def srem(self, *args, **kwargs):
        return self.client.srem(*args, **kwargs)

    @omit_exception
    def sscan(self, *args, **kwargs):
        return self.client.sscan(*args, **kwargs)

    @omit_exception
    def sscan_iter(self, *args, **kwargs):
        return self.client.sscan_iter(*args, **kwargs)

    @omit_exception
    def smismember(self, *args, **kwargs):
        return self.client.smismember(*args, **kwargs)

    @omit_exception
    def sunion(self, *args, **kwargs):
        return self.client.sunion(*args, **kwargs)

    @omit_exception
    def sunionstore(self, *args, **kwargs):
        return self.client.sunionstore(*args, **kwargs)

    @omit_exception
    def hset(self, *args, **kwargs):
        return self.client.hset(*args, **kwargs)

    @omit_exception
    def hdel(self, *args, **kwargs):
        return self.client.hdel(*args, **kwargs)

    @omit_exception
    def hlen(self, *args, **kwargs):
        return self.client.hlen(*args, **kwargs)

    @omit_exception
    def hkeys(self, *args, **kwargs):
        return self.client.hkeys(*args, **kwargs)

    @omit_exception
    def hexists(self, *args, **kwargs):
        return self.client.hexists(*args, **kwargs)



import platform
import zlib

from django.conf import settings

from django_valkey.compressors.base import BaseCompressor


class ZlibCompressor(BaseCompressor):
    """
    zlib compression
    set with:
    CACHES = {
        "default": {
            # ...
            "OPTIONS": {
                "COMPRESSOR": "django_valkey.compressors.zlib.ZlibCompressor",
                }
            }
        }

    compression parameters:
    to set `level` use `CACHE_COMPRESS_LEVEL` in your settings, defaults to 4.
    to set `minimum size` set `CACHE_COMPRESS_MIN_LENGTH` in your settings, defaults to 15.

    to set `wbits` use `COMPRESS_ZLIB_WBITS` in your settings, defaults to 15.
    this works for both compression and decompression

    """

    wbits = getattr(settings, "COMPRESS_ZLIB_WBITS", 15)

    def _compress(self, value: bytes) -> bytes:
        if int(platform.python_version_tuple()[1]) >= 11:
            return zlib.compress(value, level=self.level or 6, wbits=self.wbits)
        else:
            return zlib.compress(value, level=self.level or 6)

    def _decompress(self, value: bytes) -> bytes:
        return zlib.decompress(value, wbits=self.wbits)



import contextlib

from valkey.asyncio.client import Valkey as AValkey

from django_valkey.base import BaseValkeyCache
from django_valkey.cache import omit_exception, CONNECTION_INTERRUPTED
from django_valkey.async_cache.client.default import AsyncDefaultClient


class AsyncValkeyCache(BaseValkeyCache[AsyncDefaultClient, AValkey]):
    DEFAULT_CLIENT_CLASS = "django_valkey.async_cache.client.default.AsyncDefaultClient"

    @omit_exception
    async def set(self, *args, **kwargs):
        return await self.client.aset(*args, **kwargs)

    aset = set

    @omit_exception
    async def incr_version(self, *args, **kwargs):
        return await self.client.aincr_version(*args, **kwargs)

    aincr_version = incr_version

    @omit_exception
    async def add(self, *args, **kwargs):
        return await self.client.aadd(*args, **kwargs)

    aadd = add

    async def get(self, key, default=None, version=None, client=None):
        value = await self._get(key, default, version, client)
        if value is CONNECTION_INTERRUPTED:
            value = default
        return value

    aget = get

    @omit_exception(return_value=CONNECTION_INTERRUPTED)
    async def _get(self, key, default=None, version=None, client=None):
        return await self.client.aget(key, default, version, client)

    async def delete(self, *args, **kwargs):
        result = await self.client.adelete(*args, **kwargs)
        return bool(result)

    adelete = delete

    @omit_exception
    async def delete_pattern(self, *args, **kwargs):
        kwargs.setdefault("itersize", self._default_scan_itersize)
        return await self.client.adelete_pattern(*args, **kwargs)

    adelete_pattern = delete_pattern

    @omit_exception
    async def delete_many(self, *args, **kwargs):
        return await self.client.adelete_many(*args, **kwargs)

    adelete_many = delete_many

    @omit_exception
    async def clear(self):
        return await self.client.aclear()

    aclear = clear

    @omit_exception(return_value={})
    async def get_many(self, *args, **kwargs):
        return await self.client.aget_many(*args, **kwargs)

    aget_many = get_many

    @omit_exception
    async def set_many(self, *args, **kwargs):
        return await self.client.aset_many(*args, **kwargs)

    aset_many = set_many

    @omit_exception
    async def incr(self, *args, **kwargs):
        return await self.client.aincr(*args, **kwargs)

    aincr = incr

    @omit_exception
    async def decr(self, *args, **kwargs):
        return await self.client.adecr(*args, **kwargs)

    adecr = decr

    @omit_exception
    async def has_key(self, *args, **kwargs):
        return await self.client.ahas_key(*args, **kwargs)

    ahas_key = has_key

    @omit_exception
    async def keys(self, *args, **kwargs):
        return await self.client.akeys(*args, **kwargs)

    akeys = keys

    @omit_exception
    async def iter_keys(self, *args, **kwargs):
        async with contextlib.aclosing(self.client.aiter_keys(*args, **kwargs)) as it:
            async for key in it:
                yield key

    aiter_keys = iter_keys

    @omit_exception
    async def ttl(self, *args, **kwargs):
        return await self.client.attl(*args, **kwargs)

    attl = ttl

    @omit_exception
    async def pttl(self, *args, **kwargs):
        return await self.client.apttl(*args, **kwargs)

    apttl = pttl

    @omit_exception
    async def persist(self, *args, **kwargs):
        return await self.client.persist(*args, **kwargs)

    apersist = persist

    @omit_exception
    async def expire(self, *args, **kwargs):
        return await self.client.expire(*args, **kwargs)

    aexpire = expire

    @omit_exception
    async def expire_at(self, *args, **kwargs):
        return await self.client.expire_at(*args, **kwargs)

    aexpire_at = expire_at

    @omit_exception
    async def pexpire(self, *args, **kwargs):
        return await self.client.pexpire(*args, **kwargs)

    apexpire = pexpire

    @omit_exception
    async def pexpire_at(self, *args, **kwargs):
        return await self.client.pexpire_at(*args, **kwargs)

    apexpire_at = pexpire_at

    @omit_exception
    async def get_lock(self, *args, **kwargs):
        return await self.client.aget_lock(*args, **kwargs)

    lock = alock = aget_lock = get_lock

    @omit_exception
    async def aclose(self, *args, **kwargs):
        return await self.client.aclose()

    @omit_exception
    async def touch(self, *args, **kwargs):
        return await self.client.touch(*args, **kwargs)

    atouch = touch

    @omit_exception
    async def sadd(self, *args, **kwargs):
        return await self.client.sadd(*args, **kwargs)

    asadd = sadd

    @omit_exception
    async def scard(self, *args, **kwargs):
        return await self.client.scard(*args, **kwargs)

    ascard = scard

    @omit_exception
    async def sdiff(self, *args, **kwargs):
        return await self.client.sdiff(*args, **kwargs)

    asdiff = sdiff

    @omit_exception
    async def sdiffstore(self, *args, **kwargs):
        return await self.client.sdiffstore(*args, **kwargs)

    asdiffstore = sdiffstore

    @omit_exception
    async def sinter(self, *args, **kwargs):
        return await self.client.sinter(*args, **kwargs)

    asinter = sinter

    @omit_exception
    async def sinterstore(self, *args, **kwargs):
        return await self.client.sinterstore(*args, **kwargs)

    asinterstore = sinterstore

    @omit_exception
    async def sismember(self, *args, **kwargs):
        return await self.client.sismember(*args, **kwargs)

    asismember = sismember

    @omit_exception
    async def smembers(self, *args, **kwargs):
        return await self.client.smembers(*args, **kwargs)

    asmembers = smembers

    @omit_exception
    async def smove(self, *args, **kwargs):
        return await self.client.smove(*args, **kwargs)

    asmove = smove

    @omit_exception
    async def spop(self, *args, **kwargs):
        return await self.client.spop(*args, **kwargs)

    aspop = spop

    @omit_exception
    async def srandmember(self, *args, **kwargs):
        return await self.client.srandmember(*args, **kwargs)

    asrandmember = srandmember

    @omit_exception
    async def srem(self, *args, **kwargs):
        return await self.client.srem(*args, **kwargs)

    asrem = srem

    @omit_exception
    async def sscan(self, *args, **kwargs):
        return await self.client.sscan(*args, **kwargs)

    asscan = sscan

    @omit_exception
    async def sscan_iter(self, *args, **kwargs):
        async with contextlib.aclosing(self.client.sscan_iter(*args, **kwargs)) as it:
            async for key in it:
                yield key

    asscan_iter = sscan_iter

    @omit_exception
    async def smismember(self, *args, **kwargs):
        return await self.client.smismember(*args, **kwargs)

    asmismember = smismember

    @omit_exception
    async def sunion(self, *args, **kwargs):
        return await self.client.sunion(*args, **kwargs)

    asunion = sunion

    @omit_exception
    async def sunionstore(self, *args, **kwargs):
        return await self.client.sunionstore(*args, **kwargs)

    asunionstore = sunionstore

    @omit_exception
    async def hset(self, *args, **kwargs):
        return await self.client.hset(*args, **kwargs)

    ahset = hset

    @omit_exception
    async def hdel(self, *args, **kwargs):
        return await self.client.hdel(*args, **kwargs)

    ahdel = hdel

    @omit_exception
    async def hlen(self, *args, **kwargs):
        return await self.client.hlen(*args, **kwargs)

    ahlen = hlen

    @omit_exception
    async def hkeys(self, *args, **kwargs):
        return await self.client.hkeys(*args, **kwargs)

    ahkeys = hkeys

    @omit_exception
    async def hexists(self, *args, **kwargs):
        return await self.client.hexists(*args, **kwargs)

    ahexists = hexists



import pyzstd

from django.conf import settings

from django_valkey.compressors.base import BaseCompressor


class ZStdCompressor(BaseCompressor):
    """
    ZStdCompressor
    set with:
    CACHES = {
        "default": {
            # ...
            "OPTIONS": {
                "COMPRESSOR": "django_valkey.compressors.zstd.ZStdCompressor",
                }
            }
        }

    compression parameters:
    to set `level_or_option` use either `CACHE_COMPRESS_LEVEL` or `COMPRESS_ZSTD_OPTIONS` in your settings.
    if `COMPRESSION_ZSTD_OPTIONS` is set, level won't be used

    to set `minimum size` set `CACHE_COMPRESS_MIN_LENGTH` in your settings, defaults to 15.

    to set `zstd_dict` use `COMPRESS_ZSTD_DICT` in your settings, it defaults to None.

    decompression parameters:
    to set `options` use `DECOMPRESS_ZSTD_OPTIONS` in your settings,
    if no value provided, it uses `COMPRESS_ZSTD_OPTIONS`.

    to set `zstd_dict` use `DECOMPRESS_ZSTD_DICT` in your settings,
    if no value is provided `COMPRESS_ZSTD_DICT` is used.

    """

    options: dict | None = getattr(settings, "COMPRESS_ZSTD_OPTIONS", None)
    decomp_options = getattr(settings, "DECOMPRESS_ZSTD_OPTIONS", None)
    zstd_dict = getattr(settings, "COMPRESS_ZSTD_DICT", None)
    decomp_zstd_dict = getattr(settings, "DECOMPRESS_ZSTD_DICT", None)

    def _compress(self, value: bytes) -> bytes:
        return pyzstd.compress(
            value,
            level_or_option=self.options or self.level or 1,
            zstd_dict=self.zstd_dict,
        )

    def _decompress(self, value: bytes) -> bytes:
        return pyzstd.decompress(
            value,
            zstd_dict=self.decomp_zstd_dict or self.zstd_dict,
            option=self.decomp_options or self.options,
        )



import brotli

from django.conf import settings

from django_valkey.compressors.base import BaseCompressor


class BrotliCompressor(BaseCompressor):
    """
     Brotli compressor
     set with:
     CACHES = {
         "default": {
             # ...
             "OPTIONS": {
                 "COMPRESSOR": "django_valkey.compressors.brotli.BrotliCompressor",
                 }
             }
         }

    compression parameters:
     to set `quality` use `CACHE_COMPRESS_LEVEL` in your settings, defaults to 4.
     to set `minimum size` set `CACHE_COMPRESS_MIN_LENGTH` in your settings, defaults to 15.
     to set `lgwin` use `COMPRESS_BROTLI_LGWIN` in your settings, defaults to 22.
     to set `lgblock` use `COMPRESS_BROTLI_LGBLOCK` in your settings, defaults to 0.

     to set `mode` use `COMPRESS_BROTLI_MODE` in your settings,
     the accepted values are: "GENERIC", "TEXT", "FONT"
     the default is GENERIC, if other value is written, GENERIC will be used instead
    """

    mode = getattr(settings, "COMPRESS_BROTLI_MODE", "GENERIC")
    lgwin = getattr(settings, "COMPRESS_BROTLI_LGWIN", 22)
    lgblock = getattr(settings, "COMPRESS_BROTLI_LGBLOCK", 0)

    match mode.upper():
        case "TEXT":
            mode = brotli.MODE_TEXT
        case "FONT":
            mode = brotli.MODE_FONT
        case _:
            mode = brotli.MODE_GENERIC

    def _compress(self, value):
        return brotli.compress(
            value,
            quality=self.level or 11,
            lgwin=self.lgwin,
            lgblock=self.lgblock,
            mode=self.mode,
        )

    def _decompress(self, value):
        return brotli.decompress(value)



from typing import Any


class BaseSerializer:
    def __init__(self, options):
        pass

    def dumps(self, value: Any) -> bytes:
        raise NotImplementedError

    def loads(self, value: bytes) -> Any:
        raise NotImplementedError



from django_valkey.compressors.base import BaseCompressor


class IdentityCompressor(BaseCompressor):
    """
    the default class used by django_valkey
    it simple returns the value, with no change
    exists to simplify switching to compressors
    """

    def _compress(self, value: bytes) -> bytes:
        return value

    def _decompress(self, value: bytes) -> bytes:
        return value



import contextlib
import random
from contextlib import suppress
from typing import Any, Iterable, Set, cast, TYPE_CHECKING, AsyncGenerator

from django.conf import settings
from django.core.cache.backends.base import DEFAULT_TIMEOUT, get_key_func
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from valkey.asyncio import Valkey as AValkey
from valkey.exceptions import ResponseError
from valkey.typing import PatternT

from django_valkey import pool
from django_valkey.client.default import special_re, _main_exceptions, glob_escape
from django_valkey.exceptions import CompressorError, ConnectionInterrupted
from django_valkey.util import CacheKey

if TYPE_CHECKING:
    from django_valkey.async_cache.cache import AsyncValkeyCache


async def glove_escape(s: str) -> str:
    return special_re.sub(r"[\1]", s)


class AsyncDefaultClient:
    def __init__(
        self,
        server: str | Iterable,
        params: dict[str, Any],
        backend: "AsyncValkeyCache",
    ) -> None:
        self._backend = backend
        self._server = server
        if not self._server:
            error_message = "Missing connections string"
            raise ImproperlyConfigured(error_message)
        if not isinstance(self._server, (list, tuple, set)):
            self._server = self._server.split(",")

        self.params = params

        self.reverse_key = get_key_func(
            params.get("REVERSE_KEY_FUNCTION")
            or "django_valkey.util.default_reverse_key"
        )

        self._clients = [None] * len(self._server)
        self._options = params.get("OPTIONS", {})
        self._replica_read_only = self._options.get("REPLICA_READ_ONLY", True)

        self._connection_factory = self._options.get(
            "DJANGO_VALKEY_CONNECTION_FACTORY",
            "django_valkey.async_cache.pool.AsyncConnectionFactory",
        )
        self.connection_factory = pool.get_connection_factory(
            options=self._options, path=self._connection_factory
        )

        serializer_path = self._options.get(
            "SERIALIZER", "django_valkey.serializers.pickle.PickleSerializer"
        )
        serializer_cls: type = import_string(serializer_path)

        compressor_path = self._options.get(
            "COMPRESSOR", "django_valkey.compressors.identity.IdentityCompressor"
        )
        compressor_cls: type = import_string(compressor_path)

        self._serializer = serializer_cls(options=self._options)
        self._compressor = compressor_cls(options=self._options)

    def __contains__(self, item) -> bool:
        c = yield from self.__contains(item)
        return c

    async def __contains(self, key) -> bool:
        yield await self.has_key(key)

    async def _has_compression_enabled(self) -> bool:
        return (
            self._options.get(
                "COMPRESSOR", "django_valkey.compressors.IdentityCompressor"
            )
            != "django_valkey.compressors.IdentityCompressor"
        )

    async def _decode_iterable_result(
        self, result: Any, convert_to_set: bool = True
    ) -> list[Any] | Set[Any] | Any | None:
        if result is None:
            return None
        if isinstance(result, list):
            if convert_to_set:
                return {await self.decode(value) for value in result}
            return [await self.decode(value) for value in result]
        return await self.decode(result)

    async def get_next_client_index(
        self, write: bool = True, tried: list[int] | None = None
    ) -> int:
        if write or len(self._server) == 1:
            return 0

        if tried is None:
            tried = []

        if tried and len(tried) < len(self._server):
            not_tried = [i for i in range(0, len(self._server)) if i not in tried]
            return random.choice(not_tried)

        return random.randint(1, len(self._server) - 1)

    async def get_client(
        self,
        write: bool = True,
        tried: list[int] | None = None,
        client: AValkey | Any | None = None,
    ) -> AValkey | Any:
        if client:
            return client

        index = await self.get_next_client_index(write=write, tried=tried)

        if self._clients[index] is None:
            self._clients[index] = await self.connect(index)

        return self._clients[index]

    async def get_client_with_index(
        self, write: bool = True, tried: list[int] | None = None
    ) -> tuple[AValkey, int]:
        index = await self.get_next_client_index(write=write, tried=tried)

        if self._clients[index] is None:
            self._clients[index] = await self.connect(index)

        return self._clients[index], index

    async def connect(self, index: int = 0) -> AValkey | Any:
        return await self.connection_factory.connect(self._server[index])

    aconnect = connect

    async def disconnect(self, index: int = 0, client=None):
        if client is None:
            client = self._clients[index]

        if client is not None:
            await self.connection_factory.disconnect(client)

    adisconnect = disconnect

    async def set(
        self,
        key,
        value,
        timeout=DEFAULT_TIMEOUT,
        version=None,
        client: AValkey | Any | None = None,
        nx=False,
        xx=False,
    ) -> bool:
        nkey = await self.make_key(key, version=version)
        nvalue = await self.encode(value)

        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        original_client = client
        tried = []

        while True:
            try:
                if client is None:
                    client, index = await self.get_client_with_index(
                        write=True, tried=tried
                    )

                if timeout is not None:
                    # convert to milliseconds
                    timeout = int(timeout) * 1000

                    if timeout <= 0:
                        if nx:
                            return not await self.has_key(
                                key, version=version, client=client
                            )

                        return bool(
                            await self.delete(key, client=client, version=version)
                        )
                return bool(await client.set(nkey, nvalue, nx=nx, px=timeout, xx=xx))

            except _main_exceptions as e:
                if (
                    not original_client
                    and not self._replica_read_only
                    and len(tried) < len(self._server)
                ):
                    tried.append(index)
                    client = None
                    continue

                raise ConnectionInterrupted(connection=client) from e

    aset = set

    async def incr_version(
        self,
        key,
        delta: int = 1,
        version: int | None = None,
        client: AValkey | None | Any = None,
    ) -> int:
        client = await self.get_client(write=True, client=client)

        new_key, old_key, value, ttl, version = await self._incr_version(
            key, delta, version, client
        )

        await self.set(new_key, value, timeout=ttl, client=client)
        await self.delete(old_key, client=client)
        return version + delta

    async def _incr_version(self, key, delta, version, client) -> tuple:
        if version is None:
            version = self._backend.version

        old_key = await self.make_key(key, version)
        value = await self.get(old_key, version=version, client=client)

        try:
            ttl = await self.ttl(old_key, version=version, client=client)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if value is None:
            error_message = f"Key '{old_key!r}' does not exist"
            raise ValueError(error_message)

        if isinstance(key, CacheKey):
            new_key = await self.make_key(
                await key.original_key(), version=version + delta
            )
        else:
            new_key = await self.make_key(key, version=version + delta)

        return new_key, old_key, value, ttl, version

    aincr_version = incr_version

    async def add(
        self,
        key,
        value,
        timeout: float | int | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> bool:
        return await self.set(
            key, value, timeout, version=version, client=client, nx=True
        )

    aadd = add

    async def get(
        self,
        key,
        default: Any | None = None,
        version: int | None = None,
        client: AValkey | None | Any = None,
    ) -> Any:
        client = await self.get_client(write=False, client=client)

        key = await self.make_key(key, version=version)

        try:
            value = await client.get(key)

        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if value is None:
            return default

        return await self.decode(value)

    aget = get

    async def persist(
        self, key, version: int | None = None, client: AValkey | Any | None = None
    ) -> bool:
        client = await self.get_client(write=True, client=client)
        key = await self.make_key(key, version=version)

        return await client.persist(key)

    apersist = persist

    async def expire(
        self,
        key,
        timeout,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> bool:
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        client = await self.get_client(write=True, client=client)

        key = await self.make_key(key, version=version)

        return await client.expire(key, timeout)

    aexpire = expire

    async def expire_at(
        self, key, when, version: int | None = None, client: AValkey | Any | None = None
    ) -> bool:
        client = await self.get_client(write=True, client=client)

        key = await self.make_key(key, version=version)

        return await client.expireat(key, when)

    aexpire_at = expire_at

    async def pexpire(
        self,
        key,
        timeout,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> bool:
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        client = await self.get_client(write=True, client=client)

        key = await self.make_key(key, version=version)

        return await client.pexpire(key, timeout)

    apexpire = pexpire

    async def pexpire_at(
        self, key, when, version: int | None = None, client: AValkey | Any | None = None
    ) -> bool:
        client = await self.get_client(write=True, client=client)

        key = await self.make_key(key, version=version)

        return await client.pexpireat(key, when)

    apexpire_at = pexpire_at

    async def get_lock(
        self,
        key,
        version: int | None = None,
        timeout: float | int | None = None,
        sleep: float = 0.1,
        blocking_timeout: float | None = None,
        client: AValkey | Any | None = None,
        thread_local: bool = True,
    ):
        """Returns a Lock object, the object then should be used in an async context manager"""

        client = await self.get_client(write=True, client=client)

        key = await self.make_key(key, version=version)

        return client.lock(
            key,
            timeout=timeout,
            sleep=sleep,
            blocking_timeout=blocking_timeout,
            thread_local=thread_local,
        )

    # TODO: delete this in future releases
    lock = alock = aget_lock = get_lock

    async def delete(
        self,
        key,
        version: int | None = None,
        prefix: str | None = None,
        client: AValkey | Any | None = None,
    ) -> int:
        client = await self.get_client(write=True, client=client)

        try:
            return await client.delete(
                await self.make_key(key, version=version, prefix=prefix)
            )
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    adelete = delete

    async def delete_pattern(
        self,
        pattern: str,
        version: int | None = None,
        prefix: str | None = None,
        client: AValkey | Any | None = None,
        itersize: int | None = None,
    ) -> int:
        """
        Remove all keys matching a pattern.
        """
        client = await self.get_client(write=True, client=client)

        pattern = await self.make_pattern(pattern, version=version, prefix=prefix)

        try:
            count = 0
            pipeline = await client.pipeline()

            async with contextlib.aclosing(
                client.scan_iter(match=pattern, count=itersize)
            ) as values:
                async for key in values:
                    await pipeline.delete(key)
                    count += 1
                await pipeline.execute()

            return count

        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    adelete_pattern = delete_pattern

    async def delete_many(
        self, keys, version: int | None = None, client: AValkey | None = None
    ) -> int:
        """Remove multiple keys at once."""
        keys = [await self.make_key(k, version=version) for k in keys]

        if not keys:
            return 0

        client = await self.get_client(write=True, client=client)

        try:
            return await client.delete(*keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    adelete_many = delete_many

    async def clear(self, client: AValkey | Any | None = None) -> None:
        """
        Flush all cache keys.
        """

        client = await self.get_client(write=True, client=client)

        try:
            await client.flushdb()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    aclear = clear

    async def decode(self, value) -> Any:
        """
        Decode the given value.
        """
        try:
            value = int(value)
        except (ValueError, TypeError):
            # Handle values that weren't compressed (small stuff)
            with suppress(CompressorError):
                value = self._compressor.decompress(value)

            value = self._serializer.loads(value)
        return value

    adecode = decode

    async def encode(self, value) -> bytes | int:
        """
        Encode the given value.
        """
        if isinstance(value, bool) or not isinstance(value, int):
            value = self._serializer.dumps(value)
            return self._compressor.compress(value)

        return value

    aencode = encode

    async def get_many(
        self, keys, version: int | None = None, client: AValkey | Any | None = None
    ) -> dict:
        if not keys:
            return {}

        client = await self.get_client(write=False, client=client)

        recovered_data = {}

        map_keys = {await self.make_key(k, version=version): k for k in keys}

        try:
            results = await client.mget(*map_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        for key, value in zip(map_keys, results):
            if value is None:
                continue
            recovered_data[map_keys[key]] = await self.decode(value)

        return recovered_data

    aget_many = get_many

    async def set_many(
        self,
        data: dict,
        timeout: float | int | None = None,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> None:
        client = await self.get_client(write=True, client=client)

        try:
            pipeline = await client.pipeline()
            for key, value in data.items():
                await self.set(key, value, timeout, version=version, client=pipeline)
            await pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    aset_many = set_many

    async def _incr(
        self,
        key,
        delta: 1,
        version: int | None = None,
        client: AValkey | Any | None = None,
        ignoree_key_check: bool = False,
    ) -> int:
        client = await self.get_client(write=True, client=client)
        key = await self.make_key(key, version=version)

        try:
            try:
                # if key expired after exists check, then we get
                # key with wrong value and ttl -1.
                # use lua script for atomicity
                if not ignoree_key_check:
                    lua = """
                    local exists = server.call('EXISTS', KEYS[1])
                    if (exists == 1) then
                        return server.call('INCRBY', KEYS[1], ARGV[1])
                    else return false end
                    """
                else:
                    lua = """
                    return server.call('INCRBY', KEYS[1], ARGV[1])
                    """
                value = await client.eval(lua, 1, key, delta)
                if value is None:
                    error_message = f"Key '{key!r}' not found"
                    raise ValueError(error_message)
            except ResponseError as e:
                # if cached value or total value is greater than 64-bit signed
                # integer.
                # elif int is encoded. so valkey sees the data as string.
                # In these situations valkey will throw ResponseError

                # try to keep TTL of key
                timeout = await self.ttl(key, version=version, client=client)

                if timeout == -2:
                    error_message = f"Key '{key!r}' not found"
                    raise ValueError(error_message) from e
                value = await self.get(key, version=version, client=client) + delta
                await self.set(
                    key, value, version=version, timeout=timeout, client=client
                )
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        return value

    async def incr(
        self,
        key,
        delta: int = 1,
        version: int | None = None,
        client: AValkey | Any | None = None,
        ignore_key_check: bool = False,
    ) -> int:
        """
        Add delta to value in the cache. If the key does not exist, raise a
        ValueError exception. if ignore_key_check=True then the key will be
        created and set to the delta value by default.
        """
        return await self._incr(
            key,
            delta,
            version=version,
            client=client,
            ignoree_key_check=ignore_key_check,
        )

    aincr = incr

    async def decr(
        self,
        key,
        delta: int = 1,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> int:
        """
        Decreace delta to value in the cache. If the key does not exist, raise a
        ValueError exception.
        """
        return await self._incr(key, delta=-delta, version=version, client=client)

    adecr = decr

    async def ttl(
        self, key, version: int | None = None, client: AValkey | Any | None = None
    ) -> int | None:
        """
        Executes TTL valkey command and return the "time-to-live" of specified key.
        If key is a non-volatile key, it returns None.
        """
        client = await self.get_client(write=False, client=client)
        key = await self.make_key(key, version=version)
        if not await client.exists(key):
            return 0

        t = await client.ttl(key)
        if t >= 0:
            return t
        if t == -2:
            return 0

        return None

    attl = ttl

    async def pttl(
        self, key, version: int | None = None, client: AValkey | Any | None = None
    ) -> int | None:
        """
        Executes PTTL valkey command and return the "time-to-live" of specified key.
        If key is a non-volatile key, it returns None.
        """
        client = await self.get_client(write=False, client=client)

        key = await self.make_key(key, version=version)
        if not await client.exists(key):
            return 0

        t = await client.pttl(key)

        if t >= 0:
            return t
        if t == -2:
            return 0

        return None

    apttl = pttl

    async def has_key(
        self, key, version: int | None = None, client: AValkey | Any | None = None
    ) -> bool:
        """
        Test if key exists.
        """
        client = await self.get_client(write=False, client=client)

        key = await self.make_key(key, version=version)
        try:
            return await client.exists(key) == 1
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    ahas_key = has_key

    async def iter_keys(
        self,
        search: str,
        itersize: int | None = None,
        client: AValkey | Any | None = None,
        version: int | None = None,
    ) -> AsyncGenerator:
        """
        Same as keys, but uses cursors
        for make memory efficient keys iteration.
        """
        client = await self.get_client(write=False, client=client)
        pattern = await self.make_pattern(search, version=version)
        async with contextlib.aclosing(
            client.scan_iter(match=pattern, count=itersize)
        ) as values:
            async for item in values:
                yield self.reverse_key(item.decode())

    aiter_keys = iter_keys

    async def keys(
        self,
        search: str,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> list[Any]:
        client = await self.get_client(write=False, client=client)

        pattern = await self.make_pattern(search, version=version)
        try:
            return [self.reverse_key(k.decode()) for k in await client.keys(pattern)]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    akeys = keys

    async def make_key(
        self, key, version: int | None = None, prefix: str | None = None
    ):
        if isinstance(key, CacheKey):
            return key

        if prefix is None:
            prefix = self._backend.key_prefix

        if version is None:
            version = self._backend.version

        return CacheKey(self._backend.key_func(key, prefix, version))

    amake_key = make_key

    async def make_pattern(
        self, pattern: str, version: int | None = None, prefix: str | None = None
    ):
        if isinstance(pattern, CacheKey):
            return pattern

        if prefix is None:
            prefix = self._backend.key_prefix
        prefix = glob_escape(prefix)

        if version is None:
            version = self._backend.version
        version_str = glob_escape(str(version))
        return CacheKey(self._backend.key_func(pattern, prefix, version_str))

    amake_pattern = make_pattern

    async def sadd(
        self,
        key,
        *values: Any,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> int:
        client = await self.get_client(write=True, client=client)
        key = await self.make_key(key, version=version)
        encoded_values = [await self.encode(value) for value in values]

        return await client.sadd(key, *encoded_values)

    asadd = sadd

    async def scard(
        self, key, version: int | None = None, client: AValkey | Any | None = None
    ) -> int:
        client = await self.get_client(write=False, client=client)
        key = await self.make_key(key, version=version)
        return await client.scard(key)

    ascard = scard

    async def sdiff(
        self, *keys, version: int | None = None, client: AValkey | Any | None = None
    ) -> Set[Any]:
        client = await self.get_client(write=False, client=client)
        nkeys = [await self.make_key(key, version=version) for key in keys]
        return {await self.decode(value) for value in await client.sdiff(*nkeys)}

    asdiff = sdiff

    async def sdiffstore(
        self,
        dest,
        *keys,
        version_dest: int | None = None,
        version_keys: int | None = None,
        client: AValkey | Any | None = None,
    ) -> int:
        client = await self.get_client(write=True, client=client)
        dest = await self.make_key(dest, version=version_dest)
        nkeys = [await self.make_key(key, version=version_keys) for key in keys]
        return await client.sdiffstore(dest, *nkeys)

    asdiffstore = sdiffstore

    async def sinter(
        self, *keys, version: int | None = None, client: AValkey | Any | None = None
    ) -> Set[Any]:
        client = await self.get_client(write=False, client=client)
        nkeys = [await self.make_key(key, version=version) for key in keys]
        return {await self.decode(value) for value in await client.sinter(*nkeys)}

    asinter = sinter

    async def sinterstore(
        self,
        dest,
        *keys,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> int:
        client = await self.get_client(write=True, client=client)
        dest = await self.make_key(dest, version=version)
        nkeys = [await self.make_key(key, version=version) for key in keys]

        return await client.sinterstore(dest, *nkeys)

    asinterstore = sinterstore

    async def smismember(
        self,
        key,
        *members: Any,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> list[bool]:
        client = await self.get_client(write=False, client=client)

        key = await self.make_key(key, version=version)
        encoded_members = [await self.encode(member) for member in members]

        return [bool(value) for value in await client.smismember(key, *encoded_members)]

    asmismember = smismember

    async def sismember(
        self,
        key,
        member: Any,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> bool:
        client = await self.get_client(write=False, client=client)

        key = await self.make_key(key, version=version)
        member = await self.encode(member)
        return bool(await client.sismember(key, member))

    asismember = sismember

    async def smembers(
        self, key, version: int | None = None, client: AValkey | Any | None = None
    ) -> Set[Any]:
        client = await self.get_client(write=False, client=client)

        key = await self.make_key(key, version=version)
        return {await self.decode(value) for value in await client.smembers(key)}

    asmembers = smembers

    async def smove(
        self,
        source,
        destination,
        member: Any,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> bool:
        client = await self.get_client(write=False, client=client)
        source = await self.make_key(source, version=version)
        destination = await self.make_key(destination, version=version)
        member = await self.encode(member)
        return await client.smove(source, destination, member)

    asmove = smove

    async def spop(
        self,
        key,
        count: int | None = None,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> Set | Any:
        client = await self.get_client(write=True, client=client)
        nkey = await self.make_key(key, version=version)
        result = await client.spop(nkey, count)
        return await self._decode_iterable_result(result)

    aspop = spop

    async def srandmember(
        self,
        key,
        count: int | None = None,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> list | Any:
        client = await self.get_client(write=False, client=client)
        key = await self.make_key(key, version=version)
        result = await client.srandmember(key, count)
        return await self._decode_iterable_result(result, convert_to_set=False)

    asrandmember = srandmember

    async def srem(
        self,
        key,
        *members,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> int:
        client = await self.get_client(write=False, client=client)

        key = await self.make_key(key, version=version)
        nmembers = [await self.encode(member) for member in members]
        return await client.srem(key, *nmembers)

    asrem = srem

    async def sscan(
        self,
        key,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> Set[Any]:
        # TODO check this is correct
        if await self._has_compression_enabled() and match:
            error_message = "Using match with compression is not supported."
            raise ValueError(error_message)

        client = await self.get_client(write=False, client=client)

        key = await self.make_key(key, version=version)
        cursor, result = await client.sscan(
            key,
            match=cast(PatternT, await self.encode(match)) if match else None,
            count=count,
        )
        return {await self.decode(value) for value in result}

    asscan = sscan

    async def sscan_iter(
        self,
        key,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ):
        if await self._has_compression_enabled() and match:
            error_message = "Using match with compression is not supported."
            raise ValueError(error_message)

        client = await self.get_client(write=False, client=client)

        key = await self.make_key(key, version=version)

        async with contextlib.aclosing(
            client.sscan_iter(
                key,
                match=cast(PatternT, await self.encode(match)) if match else None,
                count=count,
            )
        ) as values:
            async for value in values:
                yield await self.decode(value)

    asscan_iter = sscan_iter

    async def sunion(
        self,
        *keys,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> Set[Any]:
        client = await self.get_client(write=False, client=client)

        nkeys = [await self.make_key(key, version=version) for key in keys]
        return {await self.decode(value) for value in await client.sunion(*nkeys)}

    asunion = sunion

    async def sunionstore(
        self,
        destination: Any,
        *keys,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> int:
        client = await self.get_client(write=True, client=client)
        destination = await self.make_key(destination, version=version)
        encoded_keys = [await self.make_key(key, version=version) for key in keys]
        return await client.sunionstore(destination, *encoded_keys)

    asunionstore = sunionstore

    async def aclose(self) -> None:
        close_flag = self._options.get(
            "CLOSE_CONNECTION",
            getattr(settings, "DJANGO_VALKEY_CLOSE_CONNECTION", False),
        )
        if close_flag:
            await self._aclose()

    close = aclose

    async def _aclose(self) -> None:
        """
        default implementation: Override in custom client
        """
        num_clients = len(self._clients)
        for index in range(num_clients):
            # TODO: check disconnect and close
            await self.disconnect(index=index)
        self._clients = [None] * num_clients

    async def touch(
        self,
        key,
        timeout: float | int | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> bool:
        """
        Sets a new expiration for a key.
        """
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        client = await self.get_client(write=True, client=client)

        key = await self.make_key(key, version=version)
        if timeout is None:
            return bool(await client.persist(key))

        # convert timeout to milliseconds
        timeout = int(timeout * 1000)
        return bool(await client.pexpire(key, timeout))

    atouch = touch

    async def hset(
        self,
        name: str,
        key,
        value,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> int:
        """
        Sets the value of hash name at key to value.
        Returns the number of fields added to the hash.
        """
        client = await self.get_client(write=True, client=client)

        nkey = await self.make_key(key, version)
        nvalue = await self.encode(value)

        return await client.hset(name, nkey, nvalue)

    ahset = hset

    async def hdel(
        self,
        name: str,
        key,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> int:
        """
        Remove keys from hash name.
        Returns the number of fields deleted from the hash.
        """
        client = await self.get_client(write=True, client=client)
        nkey = await self.make_key(key, version=version)
        return await client.hdel(name, nkey)

    ahdel = hdel

    async def hlen(self, name: str, client: AValkey | Any | None = None) -> int:
        """
        Return the number of items in hash name.
        """
        client = await self.get_client(write=False, client=client)
        return await client.hlen(name)

    ahlen = hlen

    async def hkeys(self, name: str, client: AValkey | Any | None = None) -> list[Any]:
        client = await self.get_client(write=False, client=client)

        try:
            return [self.reverse_key(k.decode()) for k in await client.hkeys(name)]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    ahkeys = hkeys

    async def hexists(
        self,
        name: str,
        key,
        version: int | None = None,
        client: AValkey | Any | None = None,
    ) -> bool:
        """
        Return True if key exists in hash name, else False.
        """
        client = await self.get_client(write=False, client=client)
        nkey = await self.make_key(key, version=version)
        return await client.hexists(name, nkey)

    ahexists = hexists



import re
from collections import OrderedDict
from datetime import datetime, timedelta
from typing import Any, Iterator, List, Set, Dict

from valkey import Valkey
from valkey.exceptions import ConnectionError
from valkey.typing import EncodableT, KeyT

from django_valkey.client.default import DEFAULT_TIMEOUT, DefaultClient
from django_valkey.exceptions import ConnectionInterrupted
from django_valkey.hash_ring import HashRing


class ShardClient(DefaultClient):
    _findhash = re.compile(r".*\{(.*)\}.*", re.I)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not isinstance(self._server, (list, tuple, set)):
            self._server = [self._server]

        self._ring = HashRing(self._server)
        self._server_dict = self.connect()

    def get_client(self, *args, **kwargs):
        raise NotImplementedError

    def connect(self, index: int = 0) -> dict:
        connection_dict = {}
        for name in self._server:
            connection_dict[name] = self.connection_factory.connect(name)
        return connection_dict

    def get_server_name(self, _key: KeyT) -> str:
        key = str(_key)
        g = self._findhash.match(key)
        if g is not None and len(g.groups()) > 0:
            key = g.groups()[0]
        return self._ring.get_node(key)

    def get_server(self, key: KeyT):
        name = self.get_server_name(key)
        return self._server_dict[name]

    def add(
        self,
        key: KeyT,
        value: EncodableT,
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().add(
            key=key, value=value, version=version, client=client, timeout=timeout
        )

    def get(
        self,
        key: KeyT,
        default: Any | None = None,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Any:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().get(key=key, default=default, version=version, client=client)

    def get_many(
        self,
        keys: KeyT,
        version: int | None = None,
        _client: Valkey | Any | None = None,
    ) -> OrderedDict:
        if not keys:
            return OrderedDict()

        recovered_data = OrderedDict()

        new_keys = [self.make_key(key, version=version) for key in keys]
        map_keys = dict(zip(new_keys, keys))

        for key in new_keys:
            client = self.get_server(key)
            value = self.get(key=key, version=version, client=client)

            if value is None:
                continue

            recovered_data[map_keys[key]] = value
        return recovered_data

    def set(
        self,
        key: KeyT,
        value: EncodableT,
        timeout: int | float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
        nx: bool = False,
        xx: bool = False,
    ) -> bool:
        """
        Persist a value to the cache, and set an optional expiration time.
        """
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().set(
            key=key,
            value=value,
            timeout=timeout,
            version=version,
            client=client,
            nx=nx,
            xx=xx,
        )

    def set_many(
        self,
        data: Dict[KeyT, EncodableT],
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> None:
        """
        Set a bunch of values in the cache at once from a dict of key/value
        pairs. This is much more efficient than calling set() multiple times.

        If timeout is given, that timeout will be used for the key; otherwise
        the default cache timeout will be used.
        """
        for key, value in data.items():
            self.set(key, value, timeout, version=version, client=client)

    def has_key(
        self, key: KeyT, version: int | None = None, client: Valkey | Any | None = None
    ) -> bool:
        """
        Test if key exists.
        """

        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        key = self.make_key(key, version=version)
        try:
            return client.exists(key) == 1
        except ConnectionError as e:
            raise ConnectionInterrupted(connection=client) from e

    def delete(
        self,
        key: KeyT,
        version: int | None = None,
        prefix: str | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().delete(key=key, version=version, client=client)

    def ttl(
        self, key: KeyT, version: int | None = None, client: Valkey | Any | None = None
    ) -> int | None:
        """
        Executes TTL valkey command and return the "time-to-live" of specified key.
        If key is a non-volatile key, it returns None.
        """

        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().ttl(key=key, version=version, client=client)

    def pttl(
        self, key: KeyT, version: int | None = None, client: Valkey | Any | None = None
    ) -> int | None:
        """
        Executes PTTL valkey command and return the "time-to-live" of specified key
        in milliseconds. If key is a non-volatile key, it returns None.
        """

        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().pttl(key=key, version=version, client=client)

    def persist(
        self, key: KeyT, version: int | None = None, client: Valkey | Any | None = None
    ) -> bool:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().persist(key=key, version=version, client=client)

    def expire(
        self,
        key: KeyT,
        timeout: int | timedelta,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().expire(key=key, timeout=timeout, version=version, client=client)

    def pexpire(
        self,
        key: KeyT,
        timeout: int | timedelta,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().pexpire(key=key, timeout=timeout, version=version, client=client)

    def pexpire_at(
        self,
        key: KeyT,
        when: datetime | int,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        """
        Set an expiry flag on a ``key`` to ``when`` on a shard client.
        ``when`` which can be represented as an integer indicating unix
        time or a Python datetime object.
        """
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().pexpire_at(key=key, when=when, version=version, client=client)

    def expire_at(
        self,
        key: KeyT,
        when: datetime | int,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        """
        Set an expiry flag on a ``key`` to ``when`` on a shard client.
        ``when`` which can be represented as an integer indicating unix
        time or a Python datetime object.
        """
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().expire_at(key=key, when=when, version=version, client=client)

    def get_lock(
        self,
        key,
        version=None,
        timeout=None,
        sleep=0.1,
        blocking_timeout=None,
        client=None,
        thread_local=True,
    ):
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().lock(
            key,
            timeout=timeout,
            sleep=sleep,
            client=client,
            blocking_timeout=blocking_timeout,
            thread_local=thread_local,
        )

    # TODO: delete in future.
    lock = get_lock

    def delete_many(
        self, keys, version=None, _client: Valkey | Any | None = None
    ) -> int:
        """
        Remove multiple keys at once.
        """
        res = 0
        for key in [self.make_key(k, version=version) for k in keys]:
            client = self.get_server(key)
            res += self.delete(key, client=client)
        return res

    def incr_version(
        self,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        new_key, old_key, value, ttl, version = self._incr_version(
            key, delta, version, client
        )
        self.set(new_key, value, timeout=ttl, client=self.get_server(new_key))
        self.delete(old_key, client=client)
        return version + delta

    def incr(
        self,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Valkey | Any | None = None,
        **kwargs,
    ) -> int:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().incr(key=key, delta=delta, version=version, client=client)

    def decr(
        self,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().decr(key=key, delta=delta, version=version, client=client)

    def iter_keys(
        self,
        search: str,
        itersize: int | None = None,
        client: Valkey | Any | None = None,
        version: int | None = None,
    ):
        """Not Implemented"""
        error_message = "iter_keys not supported on sharded client"
        raise NotImplementedError(error_message)

    def keys(
        self,
        search: str,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> List[str]:
        pattern = self.make_pattern(search, version=version)
        keys = []
        try:
            for connection in self._server_dict.values():
                keys.extend(connection.keys(pattern))
        except ConnectionError as e:
            # FIXME: technically all clients should be passed as `connection`.
            client = self.get_server(pattern)
            raise ConnectionInterrupted(connection=client) from e

        return [self.reverse_key(k.decode()) for k in keys]

    def delete_pattern(
        self,
        pattern: str,
        version: int | None = None,
        client: Valkey | Any | None = None,
        itersize: int | None = None,
        prefix: str | None = None,
    ) -> int:
        """
        Remove all keys matching pattern.
        """
        pattern = self.make_pattern(pattern, version=version, prefix=prefix)
        kwargs = {"match": pattern}
        if itersize:
            kwargs["count"] = itersize

        keys = []
        for connection in self._server_dict.values():
            keys.extend(key for key in connection.scan_iter(**kwargs))

        res = 0
        if keys:
            for connection in self._server_dict.values():
                res += connection.delete(*keys)
        return res

    def _close(self) -> None:
        for client in self._server_dict.values():
            self.disconnect(client=client)

    def touch(
        self,
        key: KeyT,
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)

        return super().touch(key=key, timeout=timeout, version=version, client=client)

    def clear(self, client=None) -> None:
        for connection in self._server_dict.values():
            connection.flushdb()

    def sadd(
        self,
        key: KeyT,
        *values: Any,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().sadd(key, *values, version=version, client=client)

    def scard(
        self,
        key: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().scard(key=key, version=version, client=client)

    def smembers(
        self,
        key: KeyT,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set[Any]:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().smembers(key=key, version=version, client=client)

    def smove(
        self,
        source: KeyT,
        destination: KeyT,
        member: Any,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if client is None:
            source = self.make_key(source, version=version)
            client = self.get_server(source)
            destination = self.make_key(destination, version=version)

        return super().smove(
            source=source,
            destination=destination,
            member=member,
            version=version,
            client=client,
        )

    def srem(
        self,
        key: KeyT,
        *members,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> int:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().srem(key, *members, version=version, client=client)

    def sscan(
        self,
        key: KeyT,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set[Any]:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().sscan(
            key=key, match=match, count=count, version=version, client=client
        )

    def sscan_iter(
        self,
        key: KeyT,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Valkey | Any | Any | None = None,
    ) -> Iterator[Any]:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().sscan_iter(
            key=key, match=match, count=count, version=version, client=client
        )

    def srandmember(
        self,
        key: KeyT,
        count: int | None = None,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set | Any:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().srandmember(key=key, count=count, version=version, client=client)

    def sismember(
        self,
        key: KeyT,
        member: Any,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> bool:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().sismember(key, member, version=version, client=client)

    def spop(
        self,
        key: KeyT,
        count: int | None = None,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> Set | Any:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().spop(key=key, count=count, version=version, client=client)

    def smismember(
        self,
        key: KeyT,
        *members,
        version: int | None = None,
        client: Valkey | Any | None = None,
    ) -> List[bool]:
        if client is None:
            key = self.make_key(key, version=version)
            client = self.get_server(key)
        return super().smismember(key, *members, version=version, client=client)



class ConnectionInterrupted(Exception):
    def __init__(self, connection, parent=None):
        self.connection = connection

    def __str__(self) -> str:
        error_type = type(self.__cause__).__name__
        error_msg = str(self.__cause__)
        return f"Valkey {error_type}: {error_msg}"


class CompressorError(Exception):
    pass



from typing import Any

import msgpack

from django_valkey.serializers.base import BaseSerializer


class MSGPackSerializer(BaseSerializer):
    def dumps(self, value: Any) -> bytes:
        return msgpack.dumps(value)

    def loads(self, value: bytes) -> Any:
        return msgpack.loads(value, raw=False)



import json
from typing import Any

from django.core.serializers.json import DjangoJSONEncoder

from django_valkey.serializers.base import BaseSerializer


class JSONSerializer(BaseSerializer):
    encoder_class = DjangoJSONEncoder

    def dumps(self, value: Any) -> bytes:
        return json.dumps(value, cls=self.encoder_class).encode()

    def loads(self, value: bytes) -> Any:
        return json.loads(value.decode())
